function [ g, inev ] = conservativeReachTube(max_steps, accuracy, dx)
% conservativeReachTube: demonstrate sampled data reach tube is conservative.
%
%   [ g, inev ] = conservativeReachTube(steps, accuracy, dx)
%  
% The backwards reach tube is conservative in the sampled data environment,
% and this routine provides an example of that conservativeness.  In this
% particular case, the target set consists of two triangles, and the
% computed minimal reach tube fills in the whole upper half-plane except
% for small gaps near the tops of the triangles.  The actual inevitable
% collision set looks more like a tiling of diamonds.
%
%     \dot x_1 = u
%     \dot x_2 = -1
%        with input u \in { -1, +1 } and delta  = 1.
%
% There are many options in the function which can only be modifed by
% changing the file.
%
% Input Parameters:
%
%   max_steps: Positive integer.  The conservativeness of the reach tube can be
%   seen after two steps, but more can be used.  Optional.  Default = 2.
%
%   accuracy: String.  Order of accuracy of the spatial and temporal
%   approximation schemes.  Choose between 'medium' (ENO2, RK2) or
%   'veryHigh' (WENO5, RK3).  The latter shows a definite improvement in
%   the sharpness of the corners of the reach tube, but requires about four
%   times as long.  Optional.  Default = 'medium'.
%
%   dx: Positive double.  Grid cell size.  Dividing this by two will
%   multiply run time by 8; consequently, it is recommended to use higher
%   accuracy before reducing the grid size.  Optional.  Default = 0.05.
%
% Output Parameters:
%
%   g: Grid structure on which data was computed.
%
%   inev: Cell vector.  Element i holds the implicit surface function for
%   the inevitably unsafe set (including the target set) after (i-1) steps.
%   (so the initial conditions can be found in entry 0).

% Copyright 2011 Ian M. Mitchell (mitchell@cs.ubc.ca).
% This software is used, copied and distributed under the licensing 
%   agreement contained in the file LICENSE in the top directory of 
%   the distribution.
%
% Created by Ian Mitchell, 2009/11/28
% Subversion tags for version control purposes.
% $Date: 2011-11-28 13:22:59 -0800 (Mon, 28 Nov 2011) $
% $Id: conservativeReachTube.m 651 2011-11-28 21:22:59Z mitchell $

  %---------------------------------------------------------------------------
  % You will see many executable lines that are commented out.
  %   These are included to show some of the options available; modify
  %   the commenting to modify the behavior.

  start_time_overall = cputime;

  %---------------------------------------------------------------------------
  %% Make sure we can see the kernel m-files.
  run('addPathToKernel');

  %---------------------------------------------------------------------------
  %% Optional input arguments.
  if(nargin < 1)
    max_steps = 2;
  end
  if(nargin < 2)
    accuracy = 'medium';
  end
  if(nargin < 3)
    dx = 0.05;
  end
  
  %---------------------------------------------------------------------------
  %% PDE solver parameters.

  % How many intermediate plots to produce (maximum)?  Since each plot must
  % fall on a sample time boundary, the actual number will depend on t_max
  % and delta.
  plot_steps = max_steps;

  % Plot after each timestep?  Overrides t_plot.
  single_step = false;

  % Pause after each plot?
  pause_after_plot = false;

  % Delete previous plot before showing next?
  delete_last_plot = false;

  % Plot in separate subplots?  Set delete_last_plot = false in this case.
  use_subplots = 0;

  % Parameters for visualizeLevelSets.
  level = 0;
  display_type = 'contour';

  % What order of accuracy to use for the computation's spatial derivatives
  % and temporal integration?
  switch(accuracy)
    case 'veryHigh'
      derivFunc = @upwindFirstWENO5;
      integrator_func = @odeCFL3;
    case 'medium'
      derivFunc = @upwindFirstENO2;
      integrator_func = @odeCFL2;
    otherwise
      error('Unknown accuracy: %s', accuracy);
  end

  % Some integration parameters apply to any integrator.
  integrator_options = odeCFLset('factorCFL', 0.9, 'stats', 'on');
  if(single_step)
    integrator_options = odeCFLset(integrator_options, 'singleStep', 'on');
  end

  %---------------------------------------------------------------------------
  %% Optional and problem specific parameters.

  % What is the timestep between data samples?
  delta = 1.0;
  
  % What inputs should we sample?
  input_samples = [ -1, +1 ];
  num_input_samples = length(input_samples);

  % Should we only allow the reach set to grow?
  restrict_update = 1;

  %---------------------------------------------------------------------------
  %% Create the grid.  
  g.dim = 2;
  g.min = [ -5; -1.0 ];
  g.max = [ +5; +6.0 ];
  g.dx = dx;
  g.bdry = @addGhostExtrapolate;
  g = processGrid(g);

  %---------------------------------------------------------------------------
  %% Create initial conditions.
  data_upper_half = -g.xs{2};
  data_l_diagonal = shapeHyperplane(g, [ +1; +1 ], [ -0.25; 0 ]);
  data_r_diagonal = shapeHyperplane(g, [ -1; +1 ], [ +0.25; 0 ]);
  data_l_triangle = shapeIntersection(data_upper_half, data_l_diagonal);
  data_r_triangle = shapeIntersection(data_upper_half, data_r_diagonal);
  data = shapeUnion(data_l_triangle, data_r_triangle);

  % Store the initial conditions
  inev = cell(max_steps, 1);
  inev{1} = data;

  %---------------------------------------------------------------------------
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  %---------------------------------------------------------------------------
  %% Set up PDE terms that implement the dynamics.

  % Since we are sampling the inputs and there are no disturbances, we
  % can compute the purely convectional velocity field for each input
  % sample.  Remember to reverse the dynamics since we solve the PDE
  % backward in time.
  scheme_func_inev = @termConvection;
  scheme_data_inev.grid = g;
  scheme_data_inev.derivFunc = derivFunc;
  
  velocity_samples = cell(num_input_samples, 1);
  for i_input_samples = 1 : num_input_samples
    velocity_samples{i_input_samples} = { -input_samples(i_input_samples); +1 };
  end

  %---------------------------------------------------------------------------
  %% Restrict the Hamiltonian so that reachable set only grows.
  % Whatever scheme handles the dynamics must already be completely set up.
  if(restrict_update)
    innerFunc = scheme_func_inev;
    innerData = scheme_data_inev;
    clear scheme_func_inev scheme_data_inev;

    % Wrap the true Hamiltonian inside the term approximation restriction routine.
    scheme_func_inev = @termRestrictUpdate;
    scheme_data_inev.innerFunc = innerFunc;
    scheme_data_inev.innerData = innerData;
    scheme_data_inev.positive = 0;

  end

  %---------------------------------------------------------------------------
  %% Initialize Display
  f = figure; %#ok<NASGU>

  % Set up subplot parameters if necessary.
  if(use_subplots)
    rows = ceil(sqrt(plot_steps));
    cols = ceil(plot_steps / rows);
    plot_num = 1;
    subplot(rows, cols, plot_num);
  end

  h = visualizeLevelSet(g, data, display_type, level, 'target unsafe set');
  set(h, 'LineStyle', '--', 'LineWidth', 1, 'LineColor', 'blue');
  xlabel('position');  ylabel('velocity');

  hold on;
  axis(g.axis);
  daspect([ 1 1 1 ]);

  fprintf('Initialization time %g\n', cputime - start_time_overall);

  %---------------------------------------------------------------------------
  %% Prepare for the main loops for computing the inevitable collision set.
  sample_now = 0;
  
  fprintf('\nStarting computation of inevitably unsafe set.\n');
  start_time_inev = cputime;
  
  %---------------------------------------------------------------------------
  %% Loop over time samples.
  while(sample_now < max_steps)

    % Initial conditions for this time step (reshaped into the
    % form appropriate for the ODE solver call).
    y_init = data(:);
    y_end = [];

    % How far to step?
    t_span = [ 0, delta ];

    start_time_sample = cputime;

    %-----------------------------------------------------------------------
    %% Loop over input samples
    for i_input_samples = 1 : num_input_samples

      % Flow field for this input sample.
      if restrict_update
        % We need to modify the velocity field on the inner scheme_data_inev
        % structure.  This is where object orientation would help.
        scheme_data_inev.innerData.velocity = velocity_samples{i_input_samples};
      else
        % There is only one scheme_data_inev structure.
        scheme_data_inev.velocity = velocity_samples{i_input_samples}; %#ok<UNRCH>
      end

      % Take a timestep. Note that the different input samples may have
      % slightly different end times, and we are not properly adjusting
      % for that fact when we ignore the first return parameter
      [ ~, y ] = feval(integrator_func, scheme_func_inev, t_span, y_init, integrator_options, scheme_data_inev);
        
      % Outcome is the union of safe sets over all samples.
      if isempty(y_end)
        y_end = y;
      else
        y_end = max(y_end, y);
      end
      
    end

    % Get back the correctly shaped data array
    data = reshape(y_end, g.shape);

    fprintf('  Finished time sample step %d in %g seconds.\n\n', sample_now, cputime - start_time_sample);

    % Increment the number of time samples we have taken.
    sample_now = sample_now + 1;

    if(pause_after_plot)
      % Wait for last plot to be digested.
      pause;
    end

    % Delete last visualization if necessary.
    if(delete_last_plot)
      delete(h);
    end

    % Move to next subplot if necessary.
    if(use_subplots)
      plot_num = plot_num + 1;
      subplot(rows, cols, plot_num);
    end

    % Create new visualization.
    h = visualizeLevelSet(g, data, display_type, level, [ 'steps = ' num2str(sample_now) ]);
    set(h, 'LineStyle', '-', 'LineWidth', 1, 'LineColor', 'black');
    daspect([ 1 1 1 ]);

    % Store the result.
    inev{sample_now + 1} = data;
    
  end

  %---------------------------------------------------------------------------
  %% Finished computing the inevitable collision region.
  
  fprintf('\nInevitable collision computation time %g seconds\n', cputime - start_time_inev);
      
  fprintf('\nTotal computation time %g seconds\n', cputime - start_time_overall);

end


