function [ data, g, data0 ] = varyingDoubleIntegratorContinuous(t_max, accuracy)
% varyingDoubleIntegratorContinuous: double integrator reach set.
%
%   [ data, g, data0 ] = varyingDoubleIntegratorContinuous(t_max, accuracy)
%  
% This is a standard reachability calculation for the state-varying double
% integrator
%
%     \dot x_1 = x_2
%     \dot x_2 = cos(2 * pi * x_1) * u
%        with input |u| \leq U and U is some constant.
%
% There are many options in the function which can only be modifed by
% changing the file.
%
% Input Parameters:
%
%   t_max: Positive double.  Time horizon of the computation.  Optional.
%   Default = 4.0 (sufficient for convergence in the continuous time case).
%
%   accuracy: String.  Order of accuracy of the numerical integrators.
%   Options are 'medium' (ENO2, RK2) or 'veryHigh' (WENO5, RK3).  The
%   latter takes about four times as long.  Default = 'medium'.
%
% Output Parameters:
%
%   data: Implicit surface function at t_max.
%
%   g: Grid structure on which data was computed.
%
%   data0: Implicit surface function at t_0.

% Copyright 2011 Ian M. Mitchell (mitchell@cs.ubc.ca).
% This software is used, copied and distributed under the licensing 
%   agreement contained in the file LICENSE in the top directory of 
%   the distribution.
%
% Created by Ian Mitchell, 2011/11/21
% Subversion tags for version control purposes.
% $Date: 2011-12-05 21:20:02 -0800 (Mon, 05 Dec 2011) $
% $Id: varyingDoubleIntegratorContinuous.m 667 2011-12-06 05:20:02Z mitchell $

  start_time_overall = cputime;

  %---------------------------------------------------------------------------
  %% Make sure we can see the kernel m-files.
  run('addPathToKernel');
  
  %---------------------------------------------------------------------------
  %% Optional input arguments.
  if(nargin < 1)
    t_max = 4.0;
  end
  if(nargin < 2)
    accuracy = 'medium';
  end
  
  %---------------------------------------------------------------------------
  %% PDE solver parameters.

  % How many intermediate plots to produce (maximum)?  Since each plot must
  % fall on a sample time boundary, the actual number will depend on t_max
  % and delta.
  plot_steps = 9;

  % How long between plots?
  t_plot = t_max / (plot_steps - 1);

  % Plot after each timestep?  Overrides t_plot.
  single_step = false;

  % How close (relative) do we need to get to tMax to be considered finished?
  small = 100 * eps;

  % Pause after each plot?
  pause_after_plot = false;

  % Delete previous plot before showing next?
  delete_last_plot = false;

  % Plot in separate subplots?  Set delete_last_plot = false in this case.
  use_subplots = 0;

  % Parameters for visualizeLevelSets.
  level = 0;
  display_type = 'contour';

  %---------------------------------------------------------------------------
  %% Optional and problem specific parameters.

  % Maximum magnitude of the input.
  input_bound = 1.0;

  % Compute with both positive and negative velocities, or only positive?
  negative_velocities = 1;

  % Should we only allow the reach set to grow?
  restrict_update = 1;

  % How to implement the dynamics in the HJ PDE?
  dynamics_term = 'convection';
  % Could also implement as Lax-Freidrichs...

  % What order of accuracy to use for the computation's spatial derivatives
  % and temporal integration?
  switch(accuracy)
    case 'veryHigh'
      derivFunc = @upwindFirstWENO5;
      integratorFunc = @odeCFL3;
    case 'medium'
      derivFunc = @upwindFirstENO2;
      integratorFunc = @odeCFL2;
    otherwise
      error('Unknown accuracy: %s', accuracy);
  end

  % Some integration parameters apply to any integrator.
  integratorOptions = odeCFLset('factorCFL', 0.9, 'stats', 'on');
  if(single_step)
    integratorOptions = odeCFLset(integratorOptions, 'singleStep', 'on');
  end

  %---------------------------------------------------------------------------
  %% Create the grid.  
  g.dim = 2;
  if negative_velocities
    g.min = [ -5.0; -2.5 ];
  else
    g.min = [ -5.0;  0 ]; %#ok<UNRCH>
  end
  g.max = [ +5.0; +2.5 ];
  g.dx = 0.05;
  g.bdry = @addGhostExtrapolate;
  g = processGrid(g);

  %---------------------------------------------------------------------------
  %% Create initial conditions.
  data = shapeRectangleByCorners(g, [ -4.5; -2.0 ], [ +4.5, +2.0 ]);

  % Then take the complement, since the outside of the set is the actual
  % target for the backwards reachability.
  data = -data;
  % If necessary, provide the IC as an output.
  if(nargout > 2)
    data0 = data;
  end

  %---------------------------------------------------------------------------
  %% Set up PDE term that implements the dynamics.

  switch(dynamics_term)
    case 'convection'
      % The optimal feedback input for the double integrator can be determined 
      % analytically and thus the optimal dynamics can be implemented as a
      % purely convectional flow field.  Remember to reverse the dynamics
      % since we solve the PDE backward in time.
      schemeFunc = @termConvection;
      schemeData.grid = g;
      schemeData.derivFunc = derivFunc;
      schemeData.velocity = { -g.xs{2}; -(-abs(cos(2 * pi * g.xs{1})) .* sign(g.xs{2}) * input_bound) };

    %case 'LaxFriedrichs'
      % A Lax-Friedrichs implementation would be more representative of what
      % would be used for more complicated dynamics, although it is likely to
      % be more dissipative and hence less accurate.

    otherwise
      error('Unknown dynamics_term: %s', dynamics_term);
  end

  %---------------------------------------------------------------------------
  %% Restrict the Hamiltonian so that reachable set only grows.
  % Whatever scheme handles the dynamics must already be completely set up.
  if(restrict_update)
    innerFunc = schemeFunc;
    innerData = schemeData;
    clear schemeFunc schemeData;

    % Wrap the true Hamiltonian inside the term approximation restriction routine.
    schemeFunc = @termRestrictUpdate;
    schemeData.innerFunc = innerFunc;
    schemeData.innerData = innerData;
    schemeData.positive = 0;
  end

  %---------------------------------------------------------------------------
  %% Initialize Display
  f = figure; %#ok<NASGU>

  % Set up subplot parameters if necessary.
  if(use_subplots)
    rows = ceil(sqrt(plot_steps));
    cols = ceil(plot_steps / rows);
    plot_num = 1;
    subplot(rows, cols, plot_num);
  end

  h = visualizeLevelSet(g, data, display_type, level, 'target unsafe set');
  set(h, 'LineStyle', ':', 'LineWidth', 2, 'LineColor', 'r');
  xlabel('position');  ylabel('velocity');

  hold on;
  axis(g.axis);
  daspect([ 1 1 1 ]);

  fprintf('Initialization time %g\n', cputime - start_time_overall);
  
  %---------------------------------------------------------------------------
  %% Loop until t_max (subject to a little roundoff).
  t_now = 0;
  start_time_reach = cputime;
  while(t_max - t_now > small * t_max)

    % Reshape data array into column vector for ode solver call.
    y0 = data(:);

    % How far to step?
    t_span = [ t_now, min(t_max, t_now + t_plot) ];

    % Take a timestep.
    [ t y ] = feval(integratorFunc, schemeFunc, t_span, y0, integratorOptions, schemeData);
    t_now = t(end);

    % Get back the correctly shaped data array
    data = reshape(y, g.shape);

    if(pause_after_plot)
      % Wait for last plot to be digested.
      pause;
    end

    % Delete last visualization if necessary.
    if(delete_last_plot)
      delete(h);
    end

    % Move to next subplot if necessary.
    if(use_subplots)
      plot_num = plot_num + 1;
      subplot(rows, cols, plot_num);
    end

    % Create new visualization.
    h = visualizeLevelSet(g, data, display_type, level, [ 't = ' num2str(t) ]);
    set(h, 'LineStyle', ':', 'LineWidth', 1, 'LineColor', 'black');
    daspect([ 1 1 1 ]);

  end

  fprintf('Reachability execution time %g seconds\n', cputime - start_time_reach);

  %---------------------------------------------------------------------------
  %% Termination code.  

  % Make the final contour a little more visible.
  set(h, 'LineStyle', '-', 'LineWidth', 2, 'LineColor', 'magenta');
  title(sprintf('Continuous time, t_{max} = %g, u_{max} = %4.2f, accuracy %s', ...
                t_max, input_bound, accuracy));
  daspect([ 1 1 1 ]);

  fprintf('Total execution time %g seconds\n', cputime - start_time_overall);

end