function [ aggressive, permissive, g_ds ] = ...
       varyingDoubleIntegratorControl(g, ctrl, inev, unsafe, input_samples, downsample)
% varyingDoubleIntegratorControl: Generate sampled data control policy.
%
%   [ inputs, aggressive, permissive, g_ds ] = ...
%       varyingDoubleIntegratorControl(g, ctrl, inev, unsafe, input_samples, downsample)
%
% Determines the set of inputs which satisfy the switched control policies
% outlined in the ADHS submission.  The inputs are chosen from a sampled
% set of input values.  All input values which satisfy the control policy
% are returned (the policies are in general not unique).
%
% Two policies are considered in the mandatory control and inevitable sets.
% The aggressive policy chooses inputs which will maximize the value of the
% inevitable implicit surface function at the end of the next sample set
% for both sets.  The permissive policy chooses inputs which ensure that
% the value of the inevitable implicit surface function at the end of the
% next sample set does not decrease in the mandatory control set, and
% allows all possible inputs in the inevitable set.
%
% The policy in the free control set is always all possible inputs.  
% The policy in the unsafe set is always no possible inputs.
%
% Input Parameters:
%
%   g: Grid structure on which data was computed.
%
%   ctrl: Array.  Implicit surface function for the control donut set
%   (including the inevitably unsafe set and the unsafe target set).
%
%   inev: Array.  Implicit surface function for the inevitably unsafe set
%   (including the unsafe target set).
%
%   unsafe: Array.  Implicit surface function for the target unsafe set.
%
%   input_samples: Vector.  Input sample values which were considered when
%   constructing the implicit surface functions.
%
%   downsample: Positive integer.  How much should the grid be downsampled
%   by before computing the input sets?  Downsampling can significantly
%   reduce the computational time; for example, a value of two will reduce
%   computational time by a factor of four.  Optional.  Default = 1 (no
%   downsampling).
%
% Output Parameters:
%
%   aggressive: Boolean array.  Same size as the (downsampled) grid, but
%   with one extra dimension the same length as input_samples.  For each
%   grid node, this array specifies which input_samples are part of the
%   aggressive policy by setting the corresponding boolean value to true.
%
%   permissive: Boolean array.  Same size as the (downsampled) grid, but
%   with one extra dimension the same length as input_samples.  For each
%   grid node, this array specifies which input_samples are part of the
%   permissive policy by setting the corresponding boolean value to true.
%
%   g_ds: Grid structure after any requested downsampling.

% Copyright 2011 Ian M. Mitchell (mitchell@cs.ubc.ca).
% This software is used, copied and distributed under the licensing 
%   agreement contained in the file LICENSE in the top directory of 
%   the distribution.
%
% Created by Ian Mitchell, 2009/11/21
% Subversion tags for version control purposes.
% $Date: 2011-11-30 14:20:14 -0800 (Wed, 30 Nov 2011) $
% $Id: varyingDoubleIntegratorControl.m 654 2011-11-30 22:20:14Z mitchell $

  %---------------------------------------------------------------------------
  %% Make sure we can see the kernel m-files.
  run('addPathToKernel');

  %---------------------------------------------------------------------------
  %% Algorithm parameters.
  
  % How close to the maximum achievable vlaue does an input have to get in
  % order to be considered for the aggressive strategy.
  aggressive_factor = 0.01;

  % How many input samples were provided?
  num_input_samples = length(input_samples);
  
  %---------------------------------------------------------------------------
  %% Optional arguments.
  if(nargin < 6)
    downsample = 1;
  end
  
  %---------------------------------------------------------------------------
  %% Perform any necessary downsampling.
  [ g_ds, ctrl_ds, inev_ds, unsafe_ds ] = downsampleGrid(g, downsample, ctrl, inev, unsafe);
  
  %---------------------------------------------------------------------------
  %% Set up some arrays.
  
  % Linear indexes to the nodes which lie in each set.  Note that ToolboxLS
  % usuaflly uses the term "index" to refer to what Matlab calls a
  % subscript.  In this routine I will adopt the Matlab nomenclature:
  % "ind" variables are linear indexes.
  ind_free = find(ctrl_ds > 0);
  num_free = length(ind_free);
  ind_ctrl = find((ctrl_ds <= 0) & (inev_ds > 0));
  num_ctrl = length(ind_ctrl);
  ind_inev = find((inev_ds <= 0) & (unsafe_ds > 0));
  num_inev = length(ind_inev);
  %ind_unsafe = find(unsafe_ds <= 0);
  
  % Logical output arrays.  Start with every input invalid.
  aggressive = false([ g_ds.shape, num_input_samples ]);
  permissive = false([ g_ds.shape, num_input_samples ]);
  
  %---------------------------------------------------------------------------
  %% Input policy for the unsafe set.
  
  % Since every input is invalid by default, the unsafe set is already
  % done.

  %---------------------------------------------------------------------------
  %% Input policy for free control set: Every input is valid in the free control set.
  
  % Get the subscripts for the nodes in the free set.
  [ subscripts{1:g_ds.dim, 1} ] = ind2sub(g_ds.shape, ind_free);

  % For each input, find the linear indexes in the output policy arrays
  % (which have an extra dimension) corresponding to that input, and set
  % the corresponding logical value to true.
  for i = 1 : num_input_samples
    ind_input_i = sub2ind([ g_ds.shape, num_input_samples ], subscripts{:}, i * ones(num_free, 1));
    aggressive(ind_input_i) = true;
    permissive(ind_input_i) = true;
  end
  
  %---------------------------------------------------------------------------
  %% Input policy for mandatory control set.
  
  % For each node and input sample, determine what the value of the
  % implicit surface function inev is along the corresponding trajectory
  % after one sample period.  Also record the best value over all inputs.
  [ best_next_value_ctrl, next_value_ctrl ] = computeNextValue(g, inev, g_ds, ind_ctrl, input_samples);
  
  % Get the subscripts for the nodes in the mandatory control set.
  [ subscripts{1:g_ds.dim, 1} ] = ind2sub(g_ds.shape, ind_ctrl);

  % Go back through the input choices flagging those which correspond to the
  % control policy.
  curr_value_ctrl = inev_ds(ind_ctrl);
  for i = 1 : num_input_samples
    ind_input_i = sub2ind([ g_ds.shape, num_input_samples ], subscripts{:}, i * ones(num_ctrl, 1));

    % Aggressive policy allows any input for which the next value is close
    % to the maximum value.
    aggressive(ind_input_i) = (next_value_ctrl{i} >= best_next_value_ctrl - aggressive_factor);
    
    % Permissive policy allows any input for which the next value does not
    % decrease.
    permissive(ind_input_i) = (next_value_ctrl{i} >= curr_value_ctrl - aggressive_factor);

  end
  
  %---------------------------------------------------------------------------
  %% Input policy for inevitable unsafe set.

  % For each node and input sample, determine what the value of the
  % implicit surface function inev is along the corresponding trajectory
  % after one sample period.  Also record the best value over all inputs.
  [ best_next_value_inev, next_value_inev ] = computeNextValue(g, inev, g_ds, ind_inev, input_samples);
  
  % Get the subscripts for the nodes in the mandatory control set.
  [ subscripts{1:g_ds.dim, 1} ] = ind2sub(g_ds.shape, ind_inev);

  % Go back through the input choices flagging those which correspond to the
  % control policy.
  for i = 1 : num_input_samples

    ind_input_i = sub2ind([ g_ds.shape, num_input_samples ], subscripts{:}, i * ones(num_inev, 1));

    % Aggressive policy allows any input for which the next value is close
    % to the maximum value.
    aggressive(ind_input_i) = (next_value_inev{i} >= best_next_value_inev - aggressive_factor);

    % The permissive policy is to allow all inputs.
    permissive(ind_input_i) = true;
  end
  
  
end


%--------------------------------------------------------------------------
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%--------------------------------------------------------------------------
function [ best_next_value, next_value ] = computeNextValue(g, data, g_ds, ind, input_samples)
% computeNextValue: Determine value at end of a single sample step trajectory.
%
%   [ best_next_value, next_value ] = computeNextValue(g, data, g_ds, ind, input_samples)
%
% For every grid node in a specified index set, integrate a trajectory
% forward for every input sample.  At the end of that trajectory,
% interpolate over a value function.  Record the resulting value for each
% grid node and for each input, and also record the best value found for
% each grid node over all inputs.
%
% Input Parameters:
%
%   g: Grid structure for interpolation (typically without downsampling).
%
%   data: Array.  Value function defined on grid g.
%
%   g_ds: Grid structure for generating initial conditions for trajectories
%   (typically with downsampling).
%
%   ind: Vector.  Linear indexes of the nodes in g_ds from which to
%   integrated trajectories.
%
%   input_samples: Vector.  Input values to use.
%
% Output Parameters:
%
%   best_next_value: Vector or array (the same shape as the arrays in the
%   elements of input parameter index).  Contains the best interpolated
%   value over all inputs for each node in the index set.
%
%   next_value: Cell vector, same length as input_samples.  Element i (same
%   shape as best_next_value vector or array) contains the interpolated
%   value at the end of the trajectory generated by input i for each node
%   in the index set.

  num_input_samples = length(input_samples);
  
  % Get the nodes from which we will integrate.
  curr_state = cell(g_ds.dim, 1);
  for d = 1 : g_ds.dim
    curr_state{d} = g_ds.xs{d}(ind);
  end
  
  % Allocate the output parameters.
  next_value = cell(num_input_samples, 1);
  best_next_value = [];

  for i = 1 : num_input_samples
    
    % Compute trajectories from every node in this set.
    next_state = computeTrajectory(curr_state, input_samples(i));
  
    % Interpolate the value at this next state
    next_value{i} = interpn(g.xs{:}, data, next_state{:});

    % Keep track of the best value at each node.
    if isempty(best_next_value)
      best_next_value = next_value{i};
    else
      best_next_value = max(best_next_value, next_value{i});
    end
    
  end
  
end


%--------------------------------------------------------------------------
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%--------------------------------------------------------------------------
function next_state = computeTrajectory(curr_state, input)
% computeTrajectory: computes trajectories for a collection of ICs
%
%   next_state = computeTrajectory(curr_state, input)
%
% Computes trajectories starting from a collection of initial conditions
% (typically nodes in the grid) using a specified fixed input for one sample
% period.
%
% In this case the trajectories evolve according to the dynamics of the
% spatially varying double integrator and the sample period is 0.3.
%
% I suspect that it would be relatively easy to modify this routine to make
% it generic to the dynamics and state space dimension.
%
% Input Parameters:
%
%   curr_state: Cell vector.  Element i contains an array which specifies
%   the coordinate of each initial condition in dimension i.  For example,
%   to integrate forward from all nodes in a grid, curr_state = grid.xs.
%
%   input: Scalar.  The fixed input to apply to every trajectory.
%
% Output Parameters:
%
%   next_state: Cell vector the same size as curr_state.  Element i
%   contains an array (the same size as the arrays in curr_state)
%   specifying the coordinate of the final position of the trajectory in
%   dimension i.

  % Sample step size.
  delta = 0.3;

  % How many trajectories?
  num_traj = length(curr_state{1});
  
  % ODE dynamics.
  f = @(t, y) [ y(num_traj+1:end); cos(2 * pi * y(1:num_traj)) * input ];
  
  % Unpack the initial conditions from the cell vector, and pack them into
  % a single regular vector.
  y0 = [ curr_state{1}; curr_state{2} ];
  
  % Run the integration.
  soln = ode45(f, [ 0, delta ], y0);
  
  % Extract the final state from the regular vector and pack it back up
  % into a cell vector.
  y_delta = deval(soln, delta);
  next_state = { y_delta(1:num_traj); y_delta(num_traj+1:end) };
  
end